# -*- coding: utf-8 -*-
"""
ComamndLine:
    python -m wbia --tf autogen_ipynb --ipynb --db PZ_MTEST --ipynb

"""
import utool as ut

COMMENT_SPACE = ut.codeblock(
    r"""
    ### Comment space below:
    """
)

IGNOREAFTER = (
    ut.codeblock(
        """
    ---

    # Developer Diagnostics  (Safely Ignored)
    # Safely Ignore the remainder of this document
    """
    ),
    None,
)


introduction = (
    ut.codeblock(
        """
    # Introduction: Testing IBEIS on a New Species

    This document is an IPython Notebook that summarizes a preliminary test
    of IBEIS on a new species. It includes steps (a) set up and configure
    the IBEIS image analysis code for a new data set, (b) run the core of
    this code, and (c) examine the results.

    To set the stage for what we are testing here, let's jump ahead to think
    about how IBEIS will work once we have built a database from initial
    training images for a species of interest. Given this database, IBEIS
    applies the following steps to each new set of images uploaded:

    1.  Gather the images taken at the same time and location (when known)
        to form what's known as an *occurrence*. Each image set is broken up
        into one or more occurrences.

    2.  Apply a *detection* process to each image to automatically find
        animals, to draw a bounding box around each, to label the species,
        and to determine the viewpoint on the animal. Together, this
        information forms an *annotation*.

    3.  Within each occurrence find all annotations that are the same
        animal, creating what IBEIS calls an *encounter.* If there are
        pictures of, say, five different animals in the occurrence then five
        encounters should be formed. Sometimes, of course, when animals are
        seen individually only one encounter will be formed.

    4.  Assign a *name* to the animal in each encounter by matching the
        annotations in the encounter against the database of animals. When
        this determines that an animal is previously unknown, a new name
        will be created and added to the database. Information about each
        encounter is also added to the database.

    Steps 3 and 4 are different applications of an underlying identification
    process whose goal is to determine which annotations show the same
    animal. The success of identification is the key to the success of
    IBEIS's image analysis. Therefore, the experiments reported in this
    IPython notebook are designed to study how well the current
    identification algorithms work by simulating Step 4 of the image
    analysis process – the hardest step. In addition to seeing how well the
    current IBEIS algorithms work, these tests can also suggest ways to
    improve the algorithms for the new species.

    Please note that the first three sections below – with (Code) in their
    title - can be safely skipped. They were generated by the IBEIS team to
    run the results here.
    """
    ),
    None,
)

nb_init = (
    '# Notebook Initialization (Code)',
    ut.codeblock(
        r"""
    # STARTBLOCK
    {autogen_str}
    # Matplotlib stuff
    import matplotlib as mpl
    %matplotlib inline
    %load_ext autoreload
    %reload_ext autoreload
    %autoreload
    import utool as ut
    import wbia.plottool as pt
    from wbia.templates import notebook_helpers
    notebook_helpers.custom_globals()

    fix_figsize = ut.partial(pt.set_figsize, w=30, h=10, dpi=256)

    # NOTE: edit the draw_case_kw to change how results appear.
    draw_case_kw = dict(show_in_notebook=True, annot_modes=[3])
    # Use annot_modes [0,1] to see with and without matches
    # See viz_qres for annotmode details
    # draw_case_kw = dict(show_in_notebook=True, annot_modes=[0,1])

    notebook_helpers.make_cells_wider()
    # ENDBLOCK
    """
    ),
)


db_init = (
    '# Database Configuration (Code)',
    ut.codeblock(
        r"""
    # STARTBLOCK
    # Setup database specific parameter configurations
    dbdir = '{dbdir}'

    # Customize one or more of the following annotation configurations
    # See wbia/expt/annotation_configs.py for an enumeration of options
    a = [
        {annotconfig_list_body}
    ]

    # Specific query / database ids can override annotation configurations
    qaid_override = None
    daid_override = None

    # Customize one ore more of the following pipeline configurations.
    # See wbia/algo/Config.py and wbia/core_annots.py for config options
    t = [
        {pipeline_list_body}
    ]

    # TODO: programatic way of listing full set of configuration options

    # Load database for this test run
    import wbia
    wbia.expt.harness.USE_BIG_TEST_CACHE = True
    ibs = wbia.opendb(dbdir=dbdir)

    if False:
        # Set to True to see some of the available LNBNN config settings
        print(ut.repr3(ibs.new_query_request([], []).qparams.hack_lnbnn_config_trail()))
        print(ut.repr3([c.asdict() for c in ibs.depc.get_config_trail('featweight', {{}})]))

    if False:
        # Valid Annot Sampling Params
        from wbia.expt import annotation_configs
        print(ut.repr3(annotation_configs.DEFAULT_AIDCFG))
    # ENDBLOCK
    """
    ),
)


fluke_select = (
    '# Humpback Select',
    ut.codeblock(
        r"""
    # STARTBLOCK
    # Tag annotations which have been given manual notch points
    from wbia_flukematch.plugin import *  # NOQA
    ibs = wbia.opendb(defaultdb='humpbacks')
    all_aids = ibs.get_valid_aids()
    isvalid = ibs.depc_annot.get_property('Has_Notch', all_aids, 'flag')
    aid_list = ut.compress(all_aids, isvalid)
    # Tag the appropriate annots
    ibs.append_annot_case_tags(aid_list, ['hasnotch'] * len(aid_list))
    #depc = ibs.depc_annot
    #qaid_override = aid_list[0:5]
    #daid_override = aid_list[0:7]
    #print(qaid_override)
    #print(daid_override)
    # ENDBLOCK
    """
    ),
)

annot_config_info = (
    '# Annotation Config Info (Safely Ignored)',
    ut.codeblock(
        r"""
    # STARTBLOCK
    acfg_list, expanded_aids_list = wbia.expt.experiment_helpers.get_annotcfg_list(
        ibs, acfg_name_list=a, qaid_override=qaid_override,
        daid_override=daid_override, verbose=0)
    # Set use_hist=True to see specific breakdowns of properties
    wbia.expt.annotation_configs.print_acfg_list(
        acfg_list, expanded_aids_list, ibs, per_qual=True, use_hist=False)
    # ENDBLOCK
    """
    ),
)


pipe_config_info = (
    '# Pipeline Config Info (Safely Ignored)',
    ut.codeblock(
        r"""
    # STARTBLOCK
    cfgdict_list, pipecfg_list = wbia.expt.experiment_helpers.get_pipecfg_list(
        test_cfg_name_list=t, ibs=ibs)
    wbia.expt.experiment_helpers.print_pipe_configs(cfgdict_list, pipecfg_list)
    # ENDBLOCK
    """
    ),
)


dbsize_expt = (
    '# Database Size Experiment ',
    ut.codeblock(
        r"""
    # STARTBLOCK
    if True:
        test_result = wbia.run_experiment(
            e='rank_surface',
            dbdir=dbdir,
            a=['varysize_td'],
            t=['candk'])
        #test_result.print_unique_annot_config_stats()
        #test_result.print_acfg_info()
        test_result.draw_func()

    if True:
        # This test requires a little bit of relaxation to get enough data
        test_result = wbia.run_experiment(
            e='rank_surface',
            dbdir=dbdir,
            a=['varysize_tdqual:qmin_pername=3,dpername=[1,2]'],
            t=['candk'])
        #test_result.print_unique_annot_config_stats()
        #test_result.print_acfg_info()
        test_result.draw_func()
    # ENDBLOCK
    """
    ),
)


timedelta_distribution = (
    '# Result Timedelta Distribution (Safely Ignore)',
    ut.codeblock(
        r"""
    # STARTBLOCK
    test_result = wbia.run_experiment(
        e='timedelta_hist',
        dbdir=dbdir,
        a=a,
        t=t,
        qaid_override=qaid_override, daid_override=daid_override,
        truepos=True)
    test_result.draw_func()
    fix_figsize()
    # ENDBLOCK
    """
    ),
)


# latex_stats = wbia.other.dbinfo.latex_dbstats([ibs], table_position='[h]') + '\n%--'
# # print(latex_stats)
# pdf_fpath = ut.compile_latex_text(latex_stats, dpath=None, verbose=False, quiet=True, pad_stdout=False)
# pdf_fpath = ut.tail(pdf_fpath, n=2)
# print(pdf_fpath)
# from IPython.display import HTML
# HTML('<iframe src="%s" width=700 height=350></iframe>' % pdf_fpath)
# _ = wbia.other.dbinfo.get_dbinfo(ibs)
timestamp_distribution = (
    ut.codeblock(
        r"""
        # Timestamp Distribution

        We can be confident that IBEIS will be successful if the
          identification process works well on photos animals that were taken
          weeks, months or years apart.
        We’d like to be able to identify an animal even if we haven’t seen it
          for a few years.
        Therefore, our first analysis studies the time distribution of the
          images.
        The plot below shows the distribution of “Timestamps” for the images
          we are testing on.
        """
    ),
    ut.codeblock(
        r"""
        # STARTBLOCK
        # Get images of those used in the tests
        acfg_list, expanded_aids_list = wbia.expt.experiment_helpers.get_annotcfg_list(
            ibs, acfg_name_list=a, qaid_override=qaid_override,
            daid_override=daid_override, verbose=0)

        aids = ut.unique(ut.total_flatten(expanded_aids_list))
        gids = ut.unique(ibs.get_annot_gids(aids))
        # Or just get time delta of all images
        #gids = ibs.get_valid_gids()

        wbia.other.dbinfo.show_image_time_distributions(ibs, gids)
        # ENDBLOCK
        """
    ),
)

example_annotations = (
    ut.codeblock(
        """
        # Example Annotations / Detections

        The detection process is not shown in this IPython notebook. This
        process – Step 2 in the description of the introduction – is something
        we usually do not apply to a new test suite of images, largely because
        it requires a fair amount of training data to adapt to a new species.
        Shown below are example annotations. Sometimes we have to draw them
        ourselves, while other times they are given to us (along with the names
        of the animals).
        """
    ),
    ut.codeblock(
        r"""
        # STARTBLOCK
        # Get a sample of images
        #gids = ibs.get_valid_gids()
        acfg_list, expanded_aids_list = wbia.expt.experiment_helpers.get_annotcfg_list(
                    ibs, acfg_name_list=a, qaid_override=qaid_override,
                    daid_override=daid_override, verbose=0)
        aids = ut.unique(ut.flatten(ut.flatten(expanded_aids_list)))
        gids = ut.unique_ordered(ibs.get_annot_gids(aids))
        # Or just get time delta of all images
        #gids = ibs.get_valid_gids()

        aids = ibs.get_image_aids(gids)

        nAids_list = list(map(len, aids))
        gids_sorted = ut.sortedby(gids, nAids_list)[::-1]
        samplex = list(range(5))
        print(samplex)
        gids_sample = ut.take(gids_sorted, samplex)

        import wbia.viz
        for gid in ut.ProgressIter(gids_sample, lbl='drawing image'):
            wbia.viz.show_image(ibs, gid)
        # ENDBLOCK
        """
    ),
    COMMENT_SPACE,
)


example_names = (
    ut.codeblock(
        """
        # Example Name Graph:

        In a test where the names for each animal in each photo are provided, we
        can link up the annotations for each animal into what we call a *name
        graph.* (This is done without running the IBEIS image analysis code –
        just from the information provided.) We can use the name graph to spot
        potential problems and get an idea of how difficult the identification
        problem might be. In particular, an animal whose name graph has fewer
        annotations will be harder to identify. Similarly, when the name graph
        includes only one annotation from a particular viewpoint – such as the
        left flank – the animal will hard to identify if that annotation by
        itself was the one we tried to use to identify the animal.
        """
    ),
    ut.codeblock(
        r"""
        # STARTBLOCK
        from wbia.viz import viz_graph
        # TODO: use viz_graph2 instead
        acfg_list, expanded_aids_list = wbia.expt.experiment_helpers.get_annotcfg_list(
                    ibs, acfg_name_list=a, qaid_override=qaid_override,
                    daid_override=daid_override, verbose=0)
        aids = ut.unique(ut.flatten(ut.flatten(expanded_aids_list)))

        # Sample some annotations
        aids = ibs.sample_annots_general(aids, filter_kw=dict(sample_size=20, min_pername=2), verbose=False)
        # Visualize name graph
        namegraph = viz_graph.make_name_graph_interaction(ibs, aids=aids, zoom=.4)
        fix_figsize()
        # ENDBLOCK
        """
    ),
    COMMENT_SPACE,
)


#######
# CONFIG COMPARISONS
#######


per_annotation_accuracy = (
    ut.codeblock(
        """
        # Annotation Identification Accuracy: (% correct annotations)

        The first and toughest test of the IBEIS identification algorithm is to
        take each annotation by itself and attempt to identify it. This is done
        by forming a database from other the annotations in the test suite and
        running the identification algorithm against this database. Other
        annotations from the same encounter as this “query annotation” are
        excluded. This is repeated for each annotation. This is the toughest
        test of image analysis because each annotation must be matched on its
        own, without any help from other annotations in the encounter. So, for
        example, if the encounter includes annotations showing both the left and
        right flank of the animal, but no other annotations in the test suite
        show this animal’s right flank, there is no hope for success in
        identifying the animal from just the right flank annotation.

        Results are shown below in both textual form and in a bar graph. It is
        probably best to skip to the bar graph. At this point, we should make
        something clear: the identification algorithm is really a ranking
        algorithm. In other words, when matching the annotation, a score is
        generated for each named animal in the database and names (and scores)
        are returned ordered by the scores. The goal is for the correct name to
        always be ranked first, but if it is ranked near the top it is usually
        ok too because the users of IBEIS make the final decisions about the
        identifying (and therefore naming) the animals.
        """
    ),
    ut.codeblock(
        r"""
        # STARTBLOCK
        testres = wbia.run_experiment(
            e='rank_cmc',
            dbdir=dbdir, a=a, t=t, qaid_override=qaid_override, daid_override=daid_override)
        #testres.print_unique_annot_config_stats()
        _ = testres.draw_func(cdfzoom=True, draw_icon=True, labelsize=40,
                              ticksize=30, figtitlesize=30, legendsize=30,
                              linewidth=5, markersize=40)
        fix_figsize()
        # ENDBLOCK
        """
    ),
    COMMENT_SPACE,
)

per_name_accuracy = (
    ut.codeblock(
        r"""
        # Encounter Identification Accuracy (% correct names)

        The second test of the IBEIS identification algorithm – the most
        important test from the user’s perspective – is to take all the
        annotations in each encounter and together use them to determine the
        identity of the animal shown in the encounter. Once again, this is done
        by forming a database from the other annotations in the test suite. We
        perform this test because IBEIS algorithms combine information from
        matching for each annotation in an encounter to determine the final
        score for each possible name and thereby the final name ranking.

        Identifying an animal using all the annotations in an encounter almost
        always improves the scoring and ranking results. This is one reason why
        we strongly encourage users of the system to take and contribute several
        different images of each animal they encounter and not try to choose the
        “best” to contribute! This simplifies the user’s job and makes the IBEIS
        software as effective as possible.
        """
    ),
    ut.codeblock(
        r"""
        # STARTBLOCK
        testres = wbia.run_experiment(
            e='rank_cmc',
            dbdir=dbdir, a=a, t=t, group_queries=True, qaid_override=qaid_override, daid_override=daid_override)
        #testres.print_unique_annot_config_stats()
        # _ = testres.draw_func()
        # fix_figsize()
        _ = testres.draw_func(cdfzoom=True, draw_icon=False, labelsize=16,
                              ticksize=16, figtitlesize=20, legendsize=16,
                              tickwidth=2, ticklength=5, gridlinewidth=1.5,
                              gridlinestyle='-', linewidth=3, markersize=10,
                              fontweight='normal', sep=' ')

        import wbia.plottool as pt
        ax = pt.gca()
        ax.grid(True)
        gridlines = ax.get_xgridlines() + ax.get_ygridlines()

        ax.xaxis.set_tick_params(width=2, length=5)
        ax.yaxis.set_tick_params(width=2, length=5)

        for line in gridlines:
            line.set_linestyle('-')
            line.set_linewidth(1.5)

        pt.set_figsize(w=12, h=8, dpi=128)
        # ENDBLOCK
        """
    ),
    COMMENT_SPACE,
)


config_overlap = (
    '# Configuration Overlap (Safely Ignored)',
    ut.codeblock(
        r"""
    # STARTBLOCK
    # How well do different configurations compliment each other?
    testres.print_config_overlap(with_plot=True)
    # ENDBLOCK
    """
    ),
)


config_disagree_cases = (
    '# Configuration Disagreements (Safely Ignored)',
    ut.codeblock(
        r"""
    # STARTBLOCK
    # This shows individual examples where the tested configurations disagree.
    # This only works if more than one configuration was specified.
    # STARTBLOCK
    testres = wbia.run_experiment(
        e='draw_cases',
        dbdir=dbdir, a=a, t=t,
        f=[':disagree=True,index=0:3'],
        figsize=(30, 8),
        **draw_case_kw)

    _ = testres.draw_func()
    # ENDBLOCK
    # ENDBLOCK
    """
    ),
)


feat_score_sep = (
    '# Feature Correspondence Score Separation (Safely Ignored)',
    ut.codeblock(
        r"""
    # STARTBLOCK
    test_result = wbia.run_experiment(
        e='TestResult.draw_feat_scoresep',
        dbdir=dbdir,
        a=a,
        t=t,
        #disttype=['L2_sift']
    )
    #test_result.draw_feat_scoresep(f='', disttype=['L2_sift'])
    test_result.draw_feat_scoresep(f='', disttype=None)
    fix_figsize()
    # ENDBLOCK
    """
    ),
)


success_annot_scoresep = (
    '# Scores of Success Cases (Safely Ignored)',
    ut.codeblock(
        r"""
        # STARTBLOCK
        testres = wbia.run_experiment(
            e='draw_annot_scoresep',
            dbdir=dbdir, a=a, t=t,
            f=[':fail=False,min_gf_timedelta=None'],
        )
        _ = testres.draw_func()
        fix_figsize()
        # ENDBLOCK
        """
    ),
)

all_annot_scoresep = (
    '# All Score Distribution (Safely Ignored)',
    ut.codeblock(
        r"""
        # STARTBLOCK
        testres = wbia.run_experiment(
            e='scores',
            dbdir=dbdir, a=a, t=t,
            qaid_override=qaid_override, daid_override=daid_override,
            f=[':fail=None,min_gf_timedelta=None']
        )
        _ = testres.draw_func()
        fix_figsize()
        #testres.draw_taghist()()
        #fix_figsize()
        # ENDBLOCK
        """
    ),
)


view_intereseting_tags = (
    '# Interesting Tags',
    ut.codeblock(
        r"""
        # STARTBLOCK
        test_result = wbia.run_experiment(
            e='draw_cases',
            dbdir=dbdir,
            a=a,
            t=t,
            f=[':index=0:5,with_tag=interesting'],
            **draw_case_kw)
        _ = test_result.draw_func()
        # ENDBLOCK
        """
    ),
)

easy_success_cases = (
    ut.codeblock(
        """
        # Examples of Top Scoring Successes:

        The next four sections including this one show examples of Annotation
        Identifications that are successes (top ranked) and failures (not top
        ranked). In each case results are shown in rows of two side-by-side
        blocks of annotations. Each block includes two, three or four
        annotations. In a block the left or upper left is the “query
        annotation”, i.e. the one IBEIS is trying to identify. This annotation
        has a purple border around it. Next to it and (when there are four
        annotations) below it are shown a set of “other” database annotations.
        When the identification is correct, these other annotations are all from
        the same name as the query annotation. These are framed in green. When
        the identified name is incorrect these other database annotations – all
        from the incorrect name - are framed in red.

        For the top scoring successes, we show strong true positives in the
        block on the left and the next best scoring true negative in the block
        on the right. Two rows of blocks are shown for each result, with the top
        row showing just the two blocks of annotations, and the blocks in the
        bottom row show the annotations together with matching regions linked by
        a line segment between annotations. The links go from the query
        annotation to the database annotations. Usually the links are to
        multiple database annotations (for the same animal) since information
        from more than one annotation is used to make decisions. Showing these
        matches allows us to understand what the image analysis algorithm used
        to score the match.

        (Notice that sometimes you will see matches between the background
        regions of the annotations. These will be eliminated in the production
        version of IBEIS for because IBEIS learns to figure out and ignore the
        background through training examples.)
        """
    ),
    ut.codeblock(
        r"""
        # STARTBLOCK
        testres = wbia.run_experiment(
            e='draw_cases',
            dbdir=dbdir, a=a, t=t,
            f=[':fail=False,index=0:3,sortdsc=gtscore,max_pername=1'],
            # REM f=[':fail=False,index=0:3,sortdsc=gtscore,without_gf_tag=Photobomb,max_pername=1'],
            # REM f=[':fail=False,sortdsc=gtscore,without_gf_tag=Photobomb,max_pername=1'],
            figsize=(30, 8),
            **draw_case_kw)

        _ = testres.draw_func()
        # ENDBLOCK
        """
    ),
    COMMENT_SPACE,
)

hard_success_cases = (
    ut.codeblock(
        """
        # Examples of Challenging Successes:

        This section shows Annotation Identification test cases where the top
        scoring name is correct, but the score is low and close to the score of
        the next best matching name.
        """
    ),
    ut.codeblock(
        r"""
        # STARTBLOCK
        testres = wbia.run_experiment(
            e='draw_cases',
            dbdir=dbdir, a=a, t=t,
            f=[':fail=False,index=0:3,sortasc=gtscore,max_pername=1,min_gtscore=.00001'],  # hack min_gtscore for 0 scores marked as success
            # REM f=[':fail=False,index=0:3,sortdsc=gtscore,without_gf_tag=Photobomb,max_pername=1'],
            # REM f=[':fail=False,sortdsc=gtscore,without_gf_tag=Photobomb,max_pername=1'],
            figsize=(30, 8),
            **draw_case_kw)

        _ = testres.draw_func()
        # ENDBLOCK
        """
    ),
    COMMENT_SPACE,
)


# ================
# Individual Cases
# ================


# '# Cases: Failure (false neg)',
failure_type2_cases = (
    ut.codeblock(
        """
        # Examples of High-Scoring Failures:

        This section shows Annotation Identification test examples where the
        correct name is not the top scoring matching, but it scored reasonably
        high. In each case, the annotations from the incorrect matching name are
        shown in the block on the left (along with the query annotation), while
        annotations from the correct matching name are shown in the block on the
        right.
        """
    ),
    ut.codeblock(
        r"""
        # STARTBLOCK
        testres = wbia.run_experiment(
            e='draw_cases',
            dbdir=dbdir, a=a, t=t,
            f=[':fail=True,index=0:3,sortdsc=gtscore,max_pername=1'],
            figsize=(30, 8),
            **draw_case_kw)
        _ = testres.draw_func()
        # ENDBLOCK
        """
    ),
    COMMENT_SPACE,
)

failure_type1_cases = (
    ut.codeblock(
        r"""
        # Examples of Low-Scoring Failures:

        This section shows Annotation Identification test examples where the
        correct name is not the top scoring match, and the correct name scored
        quite poorly. In each case, the annotations from the incorrect matching
        name are shown in the block on the left (along with the query
        annotation), while annotations from the failed but correct matching name
        are shown in the block on the right. Such failures are most often due to
        low quality annotations, occlusions, or the query annotation showing a
        view of the animal that is not in any annotation in the database.
        """
    ),
    ut.codeblock(
        r"""
        # STARTBLOCK
        testres = wbia.run_experiment(
            e='draw_cases',
            dbdir=dbdir, a=a, t=t,
            f=[':fail=True,index=0:3,sortdsc=gfscore,max_pername=1'],
            figsize=(30, 8),
            **draw_case_kw)
        _ = testres.draw_func()
        # ENDBLOCK
        """
    ),
    COMMENT_SPACE,
)


total_failure_cases = (
    ut.codeblock(
        r"""
        # Total Failures

        Shows cases where the groundtruth was not in the top 5
        """
    ),
    ut.codeblock(
        r"""
        # STARTBLOCK
        testres = wbia.run_experiment(
            e='draw_cases',
            dbdir=dbdir, a=a, t=t,
            f=[':fail=True,index=0:3,sortasc=gtscore,max_pername=1'],
            figsize=(30, 8),
            **draw_case_kw)
        _ = testres.draw_func()
        # ENDBLOCK
        """
    ),
    COMMENT_SPACE,
)


investigate_specific_case = (
    '# Cases: Custom Investigation',
    ut.codeblock(
        r"""
        # STARTBLOCK
        test_result = wbia.run_experiment(
            e='draw_cases',
            dbdir=dbdir,
            a=[a[0] + ',qsize=1'],
            #t=t,
            t=[t[0], t[0] + ':SV=False'],
            # qaid_override=[2604],  # CHOOSE A SPECIFIC ANNOTATION
            figsize=(30, 8),
            **draw_case_kw)
        _ = test_result.draw_func()
        # ENDBLOCK
        """
    ),
)


per_encounter_stats = (
    '# Combined Annot Per Encounter Info',
    ut.codeblock(
        r"""
        # STARTBLOCK
        # Load cross validation slices
        acfg_list, expanded_aids_list = wbia.expt.experiment_helpers.get_annotcfg_list(
            ibs, acfg_name_list=a, qaid_override=qaid_override,
            daid_override=daid_override, verbose=0)

        # For each slice, find how many annotations are in each encounter
        qannots_per_enc = []
        dannots_per_enc = []
        for qaids, daids in expanded_aids_list:
            qannots = ibs.annots(qaids)
            dannots = ibs.annots(daids)
            qannots_per_enc.extend(list(ut.dict_hist(qannots.encounter_text).values()))
            dannots_per_enc.extend(list(ut.dict_hist(dannots.encounter_text).values()))
        # Aggregate encounter sizes into a histogram over each slice
        query_enc_size_hist = ut.dict_hist(qannots_per_enc)
        data_enc_size_hist = ut.dict_hist(dannots_per_enc)
        print('query_enc_size_hist = ' + ut.repr4(query_enc_size_hist))
        print('data_enc_size_hist = ' + ut.repr4(data_enc_size_hist))

        # For each slice, find how many annotations are in each encounter
        qenc_per_name = []
        denc_per_name = []
        for column, list_ in enumerate([qenc_per_name, denc_per_name]):
            for aids in ut.take_column(expanded_aids_list, column):
                annots = ibs.annots(aids)
                nid_to_encounters_ = ut.group_items(annots.encounter_text, annots.nids)
                num_encounters_pername = list(ut.dict_hist(nid_to_encounters_).values())
                list_.extend(num_encounters_pername)
        query_enc_per_name_hist = ut.dict_hist(qenc_per_name)
        data_enc_per_name_hist = ut.dict_hist(denc_per_name)
        print('Mapping from #encounters to #names with that #of encounters')
        print('query_enc_per_name_hist = ' + ut.repr4(query_enc_per_name_hist))
        print('data_enc_per_name_hist = ' + ut.repr4(data_enc_per_name_hist))

        # In general
        # aids = ibs.get_valid_aids()
        # # aids = list(set(ut.flatten(ut.flatten(expanded_aids_list))))
        # annots = ibs.annots(aids)
        # nid_to_encounters_ = ut.group_items(annots.encounter_text, annots.nids)
        # nid_to_encounters = ut.map_vals(ut.unique, nid_to_encounters_)
        # num_encounters_pername = map(len, nid_to_encounters.values())
        # print('Database wide encounters per name histogram:')
        # print(ut.dict_hist(num_encounters_pername))
        # Now find out how many encounter per individual
        """
    ),
)

per_encounter_stats = (
    '# Num Annots Per Name Histogram',
    ut.codeblock(
        r"""
        a='default:is_known=True'
        aids = wbia.testdata_aids(ibs=ibs, a=a)
        annots = ibs.annots(aids)
        nid_to_aids = ut.group_items(annots.aids, annots.nids)
        nid_to_nAids = ut.map_vals(len, nid_to_aids)
        nAids_per_name_hist = ut.dict_hist(nid_to_nAids.values())

        import wbia.plottool as pt
        xdata = list(nAids_per_name_hist.keys())
        ydata = list(nAids_per_name_hist.values())
        fig = pt.multi_plot(xdata, [ydata], kind='bar',
                            xlabel='#annotations per individual',
                            ylabel='#individuals',
                            title='Annotations per individual\n' + a)

        encounters = annots.group(annots.encounter_text)[1]
        encounter_names = [a_.nids[0] for a_ in encounters]
        nid_to_encs = ut.group_items(encounters, encounter_names)
        nid_to_nEncs = ut.map_vals(len, nid_to_encs)
        nEncs_per_name_hist = ut.dict_hist(nid_to_nEncs.values())

        import wbia.plottool as pt
        xdata = list(nEncs_per_name_hist.keys())
        ydata = list(nEncs_per_name_hist.values())
        fig = pt.multi_plot(xdata, [ydata], kind='bar',
                            xlabel='#encounters per individual',
                            ylabel='#individuals',
                            title='Encounters per individual\n' + a)
        # STARTBLOCK
        """
    ),
)


# HACK
def dataset_summary_stats_hacktest():
    """
    import wbia
    ibs = wbia.opendb('WWF_Lynx_Copy')
    # import wbia
    # ibs = wbia.opendb('WWF_Lynx')
    a = [
        'default:max_timestamp=now,minqual=good,require_timestamp=True,view=left,dcrossval_enc=1,joinme=1',
        'default:max_timestamp=now,minqual=good,require_timestamp=True,view=left,dcrossval_enc=2,joinme=2',
        'default:max_timestamp=now,minqual=good,require_timestamp=True,view=left,dcrossval_enc=3,joinme=3',
        'default:max_timestamp=now,minqual=good,require_timestamp=True,view=left,dcrossval_enc=4,joinme=4',

        # 'default:max_timestamp=now,minqual=good,require_timestamp=True,view=right,dcrossval_enc=1,joinme=1',
        # 'default:max_timestamp=now,minqual=good,require_timestamp=True,view=right,dcrossval_enc=2,joinme=2',
        # 'default:max_timestamp=now,minqual=good,require_timestamp=True,view=right,dcrossval_enc=3,joinme=3',
        # 'default:max_timestamp=now,minqual=good,require_timestamp=True,view=right,dcrossval_enc=4,joinme=4',
    ]
    acfg_list, expanded_aids_list = wbia.expt.experiment_helpers.get_annotcfg_list(
        ibs, acfg_name_list=a, verbose=0)

    expt_aids = sorted(set(ut.total_flatten(expanded_aids_list)))
    print(ut.repr2(ibs.get_annot_stats_dict(expt_aids, strkeys=True, nl=2, use_hist=True)))

    # expt_qaids = sorted(set(ut.total_flatten(ut.take_column(expanded_aids_list, 0))))
    # print(ut.repr2(ibs.get_annot_stats_dict(expt_qaids, strkeys=True, nl=2, use_hist=True)))
    """
    pass


lynx_curration_stats = (
    '# Dataset Set Curation Info (lynx)',
    ut.codeblock(
        r"""
        all_aids = ibs.get_valid_aids()
        all_n_images = len(ut.unique(ibs.annots(all_aids).gids))
        all_n_names = len(ut.unique(ibs.annots(all_aids).nids))

        a_image = ['default:max_timestamp=now,require_timestamp=True']
        _, imgfilt_aids = wbia.expt.experiment_helpers.get_annotcfg_list(ibs, acfg_name_list=a_image)
        imgs = ibs.images(set(ibs.annots(imgfilt_aids).gids))
        single_aids = [x for x in imgs.aids if len(x) == 1]
        imgfilt_aids = ut.flatten(single_aids)
        imgfilt_n_images = len(ut.unique(ibs.annots(imgfilt_aids).gids))
        imgfilt_n_names = len(ut.unique(ibs.annots(imgfilt_aids).nids))

        img_removed = all_n_images - imgfilt_n_images

        print('The original image set consists {} images taken of {} named animals.'.format(
          all_n_images, all_n_names))
        print('Controlling for time and ambiguous annots removes, {} images ({}%) resulting in in {} images and {} names'.format(
            img_removed, img_removed / all_n_images, imgfilt_n_images, imgfilt_n_names
        ))

        """
    ),
)



# -*- coding: utf-8 -*-
"""Simple script to set up testdb0 & testdb1 on the filesystem"""
from pathlib import Path
from shutil import rmtree

import click

from wbia.dbio import ingest_database
from wbia.init.sysres import get_workdir


@click.command()
@click.option('-r', '--force-replace', is_flag=True, help='replace if database exists')
def main(force_replace):
    """Initializes the testdb0 & testdb1 testing directories on the filesystem"""
    workdir = Path(get_workdir())

    dbs = {
        # <name>: <factory>
        'testdb1': lambda: ingest_database.ingest_standard_database('testdb1'),
    }

    for db in dbs:
        loc = (workdir / db).resolve()
        if loc.exists():
            if force_replace:
                rmtree(loc)
            else:
                raise RuntimeError(f'{db} already exists at {loc}, aborting')
        factory = dbs[db]
        factory()
        click.echo(f'{db} created at {loc}')


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
"""
python -c "import utool as ut; ut.write_modscript_alias('Tgen.sh', 'wbia.templates.template_generator')"  # NOQA
sh Tgen.sh --key name --invert --Tcfg with_getters=True with_setters=False --modfname manual_name_funcs  # NOQA
sh Tgen.sh --key name --invert --Tcfg with_getters=True with_setters=True --modfname manual_name_funcs --funcname-filter=sex  # NOQA

"""

# TODO: Fix this name it is too special case
import logging
import uuid
import functools
import six  # NOQA

# from six.moves import range
from wbia import constants as const
from wbia.other import ibsfuncs
import numpy as np
import vtool as vt
from wbia.control import accessor_decors, controller_inject  # NOQA
import utool as ut
from wbia.control.controller_inject import make_ibs_register_decorator
import os

print, rrr, profile = ut.inject2(__name__)
logger = logging.getLogger('wbia')


CLASS_INJECT_KEY, register_ibs_method = make_ibs_register_decorator(__name__)


register_api = controller_inject.get_wbia_flask_api(__name__)


ANNOT_ROWID = 'annot_rowid'
ANNOT_SEMANTIC_UUID = 'annot_semantic_uuid'
NAME_ROWID = 'name_rowid'

NAME_UUID = 'name_uuid'
NAME_TEXT = 'name_text'
NAME_ALIAS_TEXT = 'name_alias_text'
NAME_NOTE = 'name_note'
NAME_SEX = 'name_sex'
NAME_TEMP_FLAG = 'name_temp_flag'


def testdata_ibs(defaultdb='testdb1'):
    import wbia

    ibs = wbia.opendb(defaultdb=defaultdb)
    config2_ = None  # qreq_.qparams
    return ibs, config2_


@register_ibs_method
@accessor_decors.ider
def _get_all_known_name_rowids(ibs):
    r"""
    Returns:
        list_ (list): all nids of known animals
        (does not include unknown names)
    """
    # all_known_nids = ibs._get_all_known_lblannot_rowids(const.INDIVIDUAL_KEY)
    all_known_nids = ibs.db.get_all_rowids(const.NAME_TABLE)
    return all_known_nids


@register_ibs_method
@accessor_decors.ider
def _get_all_name_rowids(ibs):
    r"""
    Returns:
        list_ (list): all nids of known animals
        (does not include unknown names)
    """
    # all_known_nids = ibs._get_all_known_lblannot_rowids(const.INDIVIDUAL_KEY)
    all_known_nids = ibs.db.get_all_rowids(const.NAME_TABLE)
    return all_known_nids


@register_ibs_method
def _get_all_known_nids(ibs):
    r"""
    alias
    """
    return _get_all_known_name_rowids(ibs)


@register_ibs_method
@accessor_decors.adder
@register_api('/api/name/', methods=['POST'])
def add_names(ibs, name_text_list, name_uuid_list=None, name_note_list=None):
    r"""
    Adds a list of names.

    Returns:
        name_rowid_list (list): their nids

    RESTful:
        Method: POST
        URL:    /api/name/
    """
    if name_note_list is None:
        name_note_list = [''] * len(name_text_list)
    # Get random uuids
    if name_uuid_list is None:
        name_uuid_list = [uuid.uuid4() for _ in range(len(name_text_list))]
    get_rowid_from_superkey = functools.partial(
        ibs.get_name_rowids_from_text, ensure=False
    )
    superkey_paramx = (1,)
    colnames = [NAME_UUID, NAME_TEXT, NAME_NOTE]
    params_iter = list(zip(name_uuid_list, name_text_list, name_note_list))
    name_rowid_list = ibs.db.add_cleanly(
        const.NAME_TABLE, colnames, params_iter, get_rowid_from_superkey, superkey_paramx
    )
    return name_rowid_list
    # OLD WAY
    # # nid_list_ = [namenid_dict[name] for name in name_list_]
    # name_text_list_ = ibs.sanitize_name_texts(name_text_list)
    # # All names are individuals and so may safely receive the INDIVIDUAL_KEY lblannot
    # lbltype_rowid = ibs.lbltype_ids[const.INDIVIDUAL_KEY]
    # lbltype_rowid_list = [lbltype_rowid] * len(name_text_list_)
    # nid_list = ibs.add_lblannots(lbltype_rowid_list, name_text_list_, note_list)
    # # nid_list = [const.UNKNOWN_NAME_ROWID if rowid is None else rowid for rowid in nid_list]
    # return nid_list


# def init_default_speciesvalue():
#    #const.KEY_DEFAULTS[const.SPECIES_KEY]
#    note_list = ['default value']
#    # Get random uuids
#    import uuid
#    lblannot_uuid_list = [uuid.UUID('00000000-0000-0000-0000-000000000001')]
#    value_list = [const.KEY_DEFAULTS[const.SPECIES_KEY]]
#    colnames = ['species_uuid', 'species_rowid', 'species_text', 'species_note']
#    params_iter = list(zip(lblannot_uuid_list, lbltype_rowid_list, value_list, note_list))
#    get_rowid_from_superkey = ibs.get_species_rowid_from_species_text
#    superkey_paramx = (1, 2)
#    species_rowid_list = ibs.db.add_cleanly(const.SPECIES_TABLE, colnames, params_iter,
#                                            get_rowid_from_superkey, superkey_paramx)


@register_ibs_method
@register_api('/api/name/sanitize/', methods=['PUT'])
def sanitize_name_texts(ibs, name_text_list):
    r"""

    RESTful:
        Method: PUT
        URL:    /api/name/sanitize
    """
    ibsfuncs.assert_valid_names(name_text_list)
    name_text_list_ = [
        None if name_text == const.UNKNOWN else name_text for name_text in name_text_list
    ]
    return name_text_list_


@register_ibs_method
@accessor_decors.deleter
@register_api('/api/name/', methods=['DELETE'])
def delete_names(ibs, name_rowid_list, safe=True, strict=False, verbose=ut.VERBOSE):
    r"""
    deletes names from the database

    CAREFUL. YOU PROBABLY DO NOT WANT TO USE THIS
    at least ensure that no annot is associated with any of these nids

    RESTful:
        Method: DELETE
        URL:    /api/name/

    # Ignore:
    #     >>> # UNPORTED_DOCTEST
    #     >>> gpath_list = grabdata.get_test_gpaths(ndata=None)[0:4]
    #     >>> gid_list = ibs.add_images(gpath_list)
    #     >>> bbox_list = [(0, 0, 100, 100)]*len(gid_list)
    #     >>> name_list = ['a', 'b', 'a', 'd']
    #     >>> aid_list = ibs.add_annots(gid_list, bbox_list=bbox_list, name_list=name_list)
    #     >>> assert len(aid_list) != 0, "No annotations added"
    #     >>> nid_list = ibs.get_valid_nids()
    #     >>> assert len(nid_list) != 0, "No names added"
    #     >>> nid = nid_list[0]
    #     >>> assert nid is not None, "nid is None"
    #     >>> ibs.delete_names(nid)
    #     >>> all_nids = ibs.get_valid_nids()
    #     >>> assert nid not in all_nids, "NID not deleted"

    """
    if verbose:
        logger.info('[ibs] deleting %d names' % len(name_rowid_list))
    if safe:
        aids_list = ibs.get_name_aids(name_rowid_list)
        aid_list = ut.flatten(aids_list)
        if strict:
            assert (
                len(aid_list) == 0
            ), 'should not be any annots belonging to a deleted name'
        else:
            if verbose:
                logger.info(
                    '[ibs] deleting %d annots that belonged to those names'
                    % len(aid_list)
                )
            if len(aid_list) > 0:
                ibs.delete_annot_nids(aid_list)
    ibs.db.delete_rowids(const.NAME_TABLE, name_rowid_list)
    # return len(name_rowid_list)
    # ibs.delete_lblannots(nid_list)


@register_ibs_method
@accessor_decors.ider
# @register_api('/api/name/nids/empty/', methods=['GET'])
def get_empty_nids(ibs, _nid_list=None):
    r"""
    get name rowids that do not have any annotations (not including UNKONWN)

    Returns:
        list: nid_list - all names without any animals (does not include unknown names)
        an nid is not invalid if it has a valid alias

    CommandLine:
        python -m wbia.control.manual_name_funcs --test-get_empty_nids

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.control.manual_name_funcs import *  # NOQA
        >>> import wbia
        >>> ibs = wbia.opendb('testdb1')
        >>> new_nid_list = ibs.make_next_nids(num=2)
        >>> empty_nids = ibs.get_empty_nids()
        >>> assert len(empty_nids) == 2, 'get_empty_nids fails1'
        >>> assert new_nid_list == empty_nids, 'get_empty_nids fails2'
        >>> ibs.delete_empty_nids()
        >>> empty_nids2 = ibs.get_empty_nids()
        >>> assert len(empty_nids2) == 0, 'get_empty_nids fails3'
        >>> result = str(empty_nids2)
        >>> print(result)
        []
    """
    recursive = _nid_list is not None
    recstr = '\t\t' if recursive else '\t'
    if _nid_list is None:
        _nid_list = ibs._get_all_known_name_rowids()
    if len(_nid_list) == 0:
        return []
    args = (len(_nid_list),)
    if recursive:
        logger.info('\tCHECKING %d NIDS FOR EMPTY (RECURSIVE)' % args)
    else:
        logger.info('CHECKING %d NIDS FOR EMPTY' % args)
    nRois_list = ibs.get_name_num_annotations(_nid_list)
    # Filter names with rois
    isempty_list = (nRois <= 0 for nRois in nRois_list)
    empty_nid_list = list(ut.iter_compress(_nid_list, isempty_list))
    # Filter names with aliases (TODO: use transitivity to determine validity)
    alias_text_list = ibs.get_name_alias_texts(empty_nid_list)
    hasalias_list = [alias_text is not None for alias_text in alias_text_list]
    # Find nids with aliases and without alias
    alias_nid_list = list(ut.ifilter_items(empty_nid_list, hasalias_list))
    no_alias_nid_list = list(ut.ifilterfalse_items(empty_nid_list, hasalias_list))
    # Find name texts and then nids of the original nids that have valid aliases
    alias_text_list = ibs.get_name_alias_texts(alias_nid_list)
    alias_nid_list = ibs.get_name_rowids_from_text(alias_text_list)
    # Find the empty aliases, recursively
    logger.info('%sFound %d empty NIDs' % (recstr, len(empty_nid_list)))
    logger.info(
        '%sFound %d empty NIDs without an alias' % (recstr, len(no_alias_nid_list))
    )
    message = ' checking these recursively' if len(alias_nid_list) > 0 else ''
    logger.info(
        '%sFound %d empty NIDs with an alias...%s'
        % (recstr, len(alias_nid_list), message)
    )
    empty_alias_nid_list = ibs.get_empty_nids(_nid_list=alias_nid_list)
    # Compile the full list of nids without any associated annotations
    empty_nid_list = empty_nid_list + no_alias_nid_list + empty_alias_nid_list
    if not recursive:
        logger.info(
            '\tFound %d empty NIDs with an alias that is recursively empty'
            % (len(empty_alias_nid_list),)
        )
    empty_nid_list = list(set(empty_nid_list))
    # Sanity check
    nRois_list = ibs.get_name_num_annotations(empty_nid_list)
    isempty_list = [nRois <= 0 for nRois in nRois_list]
    assert isempty_list.count(True) == len(empty_nid_list)
    return empty_nid_list


@register_ibs_method
# @register_api('/api/name/nids/empty/', methods=['DELETE'])
def delete_empty_nids(ibs):
    r"""
    Removes names that have no Rois from the database
    """
    logger.info('[ibs] deleting empty nids')
    invalid_nids = ibs.get_empty_nids()
    logger.info('[ibs] ... %d empty nids' % (len(invalid_nids),))
    ibs.delete_names(invalid_nids)


@register_ibs_method
@accessor_decors.getter_1toM
@register_api('/api/name/annot/rowid/', methods=['GET'])
@profile
def get_name_aids(ibs, nid_list, enable_unknown_fix=True, is_staged=False):
    r"""
    # TODO: Rename to get_anot_rowids_from_name_rowid

    Returns:
         list: aids_list a list of list of aids in each name

    RESTful:
        Method: GET
        URL:    /api/name/annot/rowid/

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.control.manual_name_funcs import *  # NOQA
        >>> import wbia
        >>> ibs = wbia.opendb('testdb1')
        >>> # Map annotations to name ids
        >>> aid_list = ibs.get_valid_aids()
        >>> nid_list = ibs.get_annot_name_rowids(aid_list)
        >>> # Get annotation ids for each name
        >>> aids_list = ibs.get_name_aids(nid_list)
        >>> # Run Assertion Test
        >>> groupid2_items = ut.group_items(aids_list, nid_list)
        >>> grouped_items = list(six.itervalues(groupid2_items))
        >>> passed_iter = map(ut.allsame, grouped_items)
        >>> passed_list = list(passed_iter)
        >>> assert all(passed_list), 'problem in get_name_aids'
        >>> # Print gropued items
        >>> print(ut.repr2(groupid2_items, newlines=False))

    Ignore;
        from wbia.control.manual_name_funcs import *  # NOQA
        import wbia
        #ibs = wbia.opendb('testdb1')
        #ibs = wbia.opendb('PZ_MTEST')
        ibs = wbia.opendb('PZ_Master0')
        #ibs = wbia.opendb('GZ_ALL')

        nid_list = ibs.get_valid_nids()
        nid_list_ = [const.UNKNOWN_NAME_ROWID if nid <= 0 else nid for nid in nid_list]

        with ut.Timer('sql'):
            #aids_list1 = ibs.get_name_aids(nid_list, enable_unknown_fix=False)
            aids_list1 = ibs.db.get(const.ANNOTATION_TABLE, (ANNOT_ROWID,), nid_list_, id_colname=NAME_ROWID, unpack_scalars=False)

        with ut.Timer('hackquery + group'):
            opstr = '''
            SELECT annot_rowid, name_rowid
            FROM annotations
            WHERE name_rowid IN
                (%s)
                ORDER BY name_rowid ASC, annot_rowid ASC
            ''' % (', '.join(map(str, nid_list)))
            pair_list = ibs.db.connection.execute(opstr).fetchall()
            aids = np.array(ut.get_list_column(pair_list, 0))
            nids = np.array(ut.get_list_column(pair_list, 1))
            unique_nids, groupx = vt.group_indices(nids)
            grouped_aids_ = vt.apply_grouping(aids, groupx)
            aids_list5 = [sorted(arr.tolist()) for arr in grouped_aids_]

        for aids1, aids5 in zip(aids_list1, aids_list5):
            if (aids1) != (aids5):
                logger.info(aids1)
                logger.info(aids5)
                logger.info('-----')

        ut.assert_lists_eq(list(map(tuple, aids_list5)), list(map(tuple, aids_list1)))

        with ut.Timer('numpy'):
            # alt method
            valid_aids = np.array(ibs.get_valid_aids())
            valid_nids = np.array(ibs.get_annot_name_rowids(valid_aids, distinguish_unknowns=False))
            aids_list2 = [valid_aids.take(np.flatnonzero(valid_nids == nid)).tolist() for nid in nid_list_]

        with ut.Timer('numpy2'):
            # alt method
            valid_aids = np.array(ibs.get_valid_aids())
            valid_nids = np.array(ibs.get_annot_name_rowids(valid_aids, distinguish_unknowns=False))
            aids_list3 = [valid_aids.take(np.flatnonzero(np.equal(valid_nids, nid))).tolist() for nid in nid_list_]

        with ut.Timer('numpy3'):
            # alt method
            valid_aids = np.array(ibs.get_valid_aids())
            valid_nids = np.array(ibs.db.get_all_col_rows(const.ANNOTATION_TABLE, NAME_ROWID))
            aids_list4 = [valid_aids.take(np.flatnonzero(np.equal(valid_nids, nid))).tolist() for nid in nid_list_]
        assert aids_list2 == aids_list3
        assert aids_list3 == aids_list4
        assert aids_list1 == aids_list2

        valid_aids = ibs.get_valid_aids()
        %timeit ibs.db.get_all_col_rows('annotations', 'rowid')
        %timeit ibs.db.get_all_col_rows('annotations', 'name_rowid')
        %timeit ibs.get_annot_name_rowids(valid_aids, distinguish_unknowns=False)
        %timeit ibs.get_valid_aids()
        %timeit ibs.get_annot_name_rowids(ibs.get_valid_aids(), distinguish_unknowns=False)
        valid_nids1 = ibs.get_annot_name_rowids(valid_aids, distinguish_unknowns=False)
        valid_nids2 = ibs.db.get_all_col_rows('annotations', 'name_rowid')
        assert valid_nids1 == valid_nids2

    ibs.db.fname
    ibs.db.fpath

    import sqlite3

    con = sqlite3.connect(ibs.db.fpath)

    opstr = '''
    SELECT annot_rowid, name_rowid
    FROM annotations
    WHERE name_rowid IN
        (SELECT name_rowid FROM name)
        ORDER BY name_rowid ASC, annot_rowid ASC
    '''

    annot_rowid_list = con.execute(opstr).fetchall()
    aid_list = ut.get_list_column(annot_rowid_list, 0)
    nid_list = ut.get_list_column(annot_rowid_list, 1)


    # HACKY HACKY HACK

    with ut.Timer('hackquery + group'):
        #nid_list = ibs.get_valid_nids()[10:15]
        nid_list = ibs.get_valid_nids()
        opstr = '''
        SELECT annot_rowid, name_rowid
        FROM annotations
        WHERE name_rowid IN
            (%s)
            ORDER BY name_rowid ASC, annot_rowid ASC
        ''' % (', '.join(map(str, nid_list)))
        pair_list = ibs.db.connection.execute(opstr).fetchall()
        aids = np.array(ut.get_list_column(pair_list, 0))
        nids = np.array(ut.get_list_column(pair_list, 1))
        unique_nids, groupx = vt.group_indices(nids)
        grouped_aids_ = vt.apply_grouping(aids, groupx)
        grouped_aids = [arr.tolist() for arr in grouped_aids_]

    SELECT
       name_rowid, COUNT(annot_rowid) AS number, GROUP_CONCAT(annot_rowid) AS aid_list
    FROM annotations
    WHERE name_rowid in (SELECT name_rowid FROM name)
     GROUP BY name_rowid
    ORDER BY name_rowid ASC


    import vtool as vt
    vt
    vt.aid_list[0]


    annot_rowid_list = con.execute(opstr).fetchall()
    opstr = '''
        SELECT annot_rowid
        FROM annotations
        WHERE name_rowid=?
        '''

    cur = ibs.db.connection.cursor()

    cur = con.execute('BEGIN IMMEDIATE TRANSACTION')
    cur = ibs.db.connection
    res = [cur.execute(opstr, (nid,)).fetchall() for nid in nid_list_]
    cur.execute('COMMIT TRANSACTION')

    res = [ibs.db.cur.execute(opstr, (nid,)).fetchall() for nid in nid_list_]

    """
    # FIXME: THIS FUNCTION IS VERY SLOW
    # ADD A LOCAL CACHE TO FIX THIS SPEED
    # ALSO FIX GET_IMAGE_AIDS
    # really a getter for the annotation table not the name table
    # return [[] for nid in nid_list]
    # TODO: should a query of the UNKNOWN_NAME_ROWID return anything?
    # TODO: don't even run negative aids as queries
    nid_list_ = [const.UNKNOWN_NAME_ROWID if nid <= 0 else nid for nid in nid_list]

    NEW_INDEX_HACK = True
    USE_GROUPING_HACK = False
    if NEW_INDEX_HACK:
        # FIXME: This index should when the database is defined.
        # Ensure that an index exists on the image column of the annotation table
        # logger.info(len(nid_list_))
        ibs.db.connection.execute(
            """
            CREATE INDEX IF NOT EXISTS nid_to_aids ON annotations (name_rowid);
            """
        ).fetchall()
        aids_list = ibs.db.get(
            const.ANNOTATION_TABLE,
            (ANNOT_ROWID,),
            nid_list_,
            id_colname=NAME_ROWID,
            unpack_scalars=False,
        )
        # %timeit ibs.db.get(const.ANNOTATION_TABLE, (ANNOT_ROWID,), nid_list_, id_colname=NAME_ROWID, unpack_scalars=False)
        # The index maxes the following query very efficient
    elif USE_GROUPING_HACK:
        # This code doesn't work because it doesn't respect empty names
        input_list, inverse_unique = np.unique(nid_list_, return_inverse=True)
        input_str = ', '.join(list(map(str, input_list)))
        opstr = """
        SELECT annot_rowid, name_rowid
        FROM {ANNOTATION_TABLE}
        WHERE name_rowid IN
            ({input_str})
            ORDER BY name_rowid ASC, annot_rowid ASC
        """.format(
            input_str=input_str, ANNOTATION_TABLE=const.ANNOTATION_TABLE
        )
        pair_list = ibs.db.connection.execute(opstr).fetchall()
        aidscol = np.array(ut.get_list_column(pair_list, 0))
        nidscol = np.array(ut.get_list_column(pair_list, 1))
        unique_nids, groupx = vt.group_indices(nidscol)
        grouped_aids_ = vt.apply_grouping(aidscol, groupx)
        # aids_list = [sorted(arr.tolist()) for arr in grouped_aids_]
        structured_aids_list = [arr.tolist() for arr in grouped_aids_]
        aids_list = np.array(structured_aids_list)[inverse_unique].tolist()
    else:
        USE_NUMPY_IMPL = True
        # USE_NUMPY_IMPL = False
        # Use qt if getting one at a time otherwise perform bulk operation
        USE_NUMPY_IMPL = len(nid_list_) > 1
        # USE_NUMPY_IMPL = len(nid_list_) > 10
        if USE_NUMPY_IMPL:
            # This seems to be 30x faster for bigger inputs
            valid_aids = np.array(ibs._get_all_aids())
            valid_nids = np.array(
                ibs.db.get_all_col_rows(const.ANNOTATION_TABLE, NAME_ROWID)
            )
            # np.array(ibs.get_annot_name_rowids(valid_aids, distinguish_unknowns=False))

            # MEMORY HOG LIKE A SON OF A BITCH
            # aids_list = [
            #     valid_aids.take(np.flatnonzero(
            #         np.equal(valid_nids, nid))).tolist()
            #     for nid in nid_list_
            # ]

            temp = np.zeros((len(valid_nids),), dtype=np.bool)
            aids_dict = {}
            nid_list_unique = np.unique(nid_list_)
            for nid in nid_list_unique:
                bool_list = np.equal(valid_nids, nid, out=temp)
                flattened = np.flatnonzero(bool_list)
                aid_list = [] if nid < 0 else valid_aids.take(flattened)
                aid_list = aid_list.tolist()
                aids_dict[nid] = aid_list

            aids_list = ut.dict_take(aids_dict, nid_list_)
        else:
            # SQL IMPL
            aids_list = ibs.db.get(
                const.ANNOTATION_TABLE,
                (ANNOT_ROWID,),
                nid_list_,
                id_colname=NAME_ROWID,
                unpack_scalars=False,
            )
    if enable_unknown_fix:
        # enable_unknown_fix == distinguish_unknowns
        # negative name rowids correspond to unknown annoations wherex annot_rowid = -name_rowid
        # aids_list = [None if nid is None else ([-nid] if nid < 0 else aids)
        #             for nid, aids in zip(nid_list, aids_list)]
        # Not sure if this should fail or return empty list on None nid
        aids_list = [
            [] if nid is None else ([-nid] if nid < 0 else aids)
            for nid, aids in zip(nid_list, aids_list)
        ]
        # aids_list = [[-nid] if nid < 0 else aids
        #             for nid, aids in zip(nid_list, aids_list)]
    aids_list = [
        ibs.filter_annotation_set(aid_list_, is_staged=is_staged)
        for aid_list_ in aids_list
    ]
    return aids_list


@register_ibs_method
@accessor_decors.getter_1toM
@register_api('/api/name/annot/uuid/', methods=['GET'])
def get_name_annot_uuids(ibs, nid_list, **kwargs):
    aids_list = ibs.get_name_aids(nid_list, **kwargs)
    annot_uuids_list = [ibs.get_annot_uuids(aid_list) for aid_list in aids_list]
    return annot_uuids_list


@register_ibs_method
@accessor_decors.getter_1toM
@register_api('/api/name/annot/rowid/exemplar/', methods=['GET'])
def get_name_exemplar_aids(ibs, nid_list):
    r"""
    Returns:
        list_ (list):  a list of list of cids in each name


    CommandLine:
        python -m wbia.control.manual_name_funcs --test-get_name_exemplar_aids

    RESTful:
        Method: GET
        URL:    /api/name/annot/rowid/examplar/

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.control.manual_name_funcs import *  # NOQA
        >>> import wbia
        >>> ibs = wbia.opendb('testdb1')
        >>> aid_list = ibs.get_valid_aids()
        >>> nid_list = ibs.get_annot_name_rowids(aid_list)
        >>> exemplar_aids_list = ibs.get_name_exemplar_aids(nid_list)
        >>> result = exemplar_aids_list
        >>> print(result)
        [[], [2, 3], [2, 3], [], [5, 6], [5, 6], [7], [8], [], [10], [], [12], [13]]
    """
    # Get all annot ids for each name
    aids_list = ibs.get_name_aids(nid_list, enable_unknown_fix=True)
    # Flag any annots that are not exemplar and remove them
    flags_list = ibsfuncs.unflat_map(ibs.get_annot_exemplar_flags, aids_list)
    exemplar_aids_list = [
        ut.compress(aids, flags) for aids, flags in zip(aids_list, flags_list)
    ]
    return exemplar_aids_list


@register_ibs_method
@accessor_decors.getter_1toM
@register_api('/api/name/annot/uuid/exemplar/', methods=['GET'])
def get_name_exemplar_name_uuids(ibs, nid_list, **kwargs):
    aids_list = ibs.get_name_exemplar_aids(nid_list, **kwargs)
    annot_uuids_list = [ibs.get_annot_uuids(aid_list) for aid_list in aids_list]
    return annot_uuids_list


@register_ibs_method
@accessor_decors.getter_1toM
@register_api('/api/name/image/rowid/', methods=['GET'])
def get_name_gids(ibs, nid_list):
    r"""
    Returns:
        list_ (list): the image ids associated with name ids

    RESTful:
        Method: GET
        URL:    /api/name/image/rowid/

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.control.manual_name_funcs import *  # NOQA
        >>> import wbia
        >>> ibs = wbia.opendb('testdb1')
        >>> nid_list = ibs._get_all_known_name_rowids()
        >>> gids_list = ibs.get_name_gids(nid_list)
        >>> result = gids_list
        >>> print(result)
        [[2, 3], [5, 6], [7], [8], [10], [12], [13]]
    """
    # TODO: Optimize
    aids_list = ibs.get_name_aids(nid_list, enable_unknown_fix=True)
    gids_list = ibsfuncs.unflat_map(ibs.get_annot_gids, aids_list)
    return gids_list


@register_ibs_method
@accessor_decors.getter_1toM
@register_api('/api/name/image/uuid/', methods=['GET'])
def get_name_image_uuids(ibs, nid_list):
    r"""
    DEPRICATE

    Returns:
        list_ (list): the image ids associated with name ids

    RESTful:
        Method: GET
        URL:    /api/name/image/uuid/
    """
    # TODO: Optimize
    gids_list = ibs.get_name_gids(nid_list)
    image_uuids_list = [ibs.get_image_uuids(gid_list) for gid_list in gids_list]
    return image_uuids_list


@register_ibs_method
@accessor_decors.getter_1to1
@register_api('/api/name/uuid/', methods=['GET'])
def get_name_uuids(ibs, nid_list):
    r"""
    Returns:
        list_ (list): uuids_list - name uuids

    RESTful:
        Method: GET
        URL:    /api/name/uuid/
    """
    uuids_list = ibs.db.get(const.NAME_TABLE, (NAME_UUID,), nid_list)
    # notes_list = ibs.get_lblannot_notes(nid_list)
    return uuids_list


@register_ibs_method
@accessor_decors.getter_1to1
@register_api('/api/name/note/', methods=['GET'])
def get_name_notes(ibs, name_rowid_list):
    r"""
    Returns:
        list_ (list): notes_list - name notes

    RESTful:
        Method: GET
        URL:    /api/name/note/
    """
    notes_list = ibs.db.get(const.NAME_TABLE, (NAME_NOTE,), name_rowid_list)
    # notes_list = ibs.get_lblannot_notes(nid_list)
    return notes_list


@register_ibs_method
@accessor_decors.getter_1to1
@register_api('/api/name/metadata/', methods=['GET'])
def get_name_metadata(ibs, name_rowid_list, return_raw=False):
    r"""
    Returns:
        list_ (list): name metadata dictionary

    RESTful:
        Method: GET
        URL:    /api/name/metadata/
    """
    metadata_str_list = ibs.db.get(
        const.NAME_TABLE, ('name_metadata_json',), name_rowid_list
    )
    metadata_list = []
    for metadata_str in metadata_str_list:
        if metadata_str in [None, '']:
            metadata_dict = {}
        else:
            metadata_dict = metadata_str if return_raw else ut.from_json(metadata_str)
        metadata_list.append(metadata_dict)
    return metadata_list


@register_ibs_method
@accessor_decors.getter_1to1
@register_api('/api/name/num/annot/', methods=['GET'])
def get_name_num_annotations(ibs, nid_list):
    r"""
    Returns:
        list_ (list):  the number of annotations for each name

    CommandLine:
        python -m wbia.control.manual_name_funcs --test-get_name_num_annotations

    RESTful:
        Method: GET
        URL:    /api/name/num/annot/

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.control.manual_name_funcs import *  # NOQA
        >>> import wbia
        >>> ibs = wbia.opendb('testdb1')
        >>> nid_list = ibs._get_all_known_name_rowids()
        >>> result = get_name_num_annotations(ibs, nid_list)
        >>> print(result)
        [2, 2, 1, 1, 1, 1, 1]
    """
    # TODO: Optimize
    return list(map(len, ibs.get_name_aids(nid_list, enable_unknown_fix=True)))


@register_ibs_method
@accessor_decors.getter_1to1
@register_api('/api/name/num/annot/exemplar/', methods=['GET'])
def get_name_num_exemplar_annotations(ibs, nid_list):
    r"""
    Returns:
        list_ (list):  the number of annotations, which are exemplars for each name

    RESTful:
        Method: GET
        URL:    /api/name/num/annot/exemplar/
    """
    return list(map(len, ibs.get_name_exemplar_aids(nid_list)))


@register_ibs_method
@register_api('/api/name/temp/', methods=['GET'])
def get_name_temp_flag(ibs, name_rowid_list, eager=True, nInput=None):
    r"""
    name_temp_flag_list <- name.name_temp_flag[name_rowid_list]

    gets data from the "native" column "name_temp_flag" in the "name" table

    Args:
        name_rowid_list (list):

    Returns:
        list: name_temp_flag_list

    TemplateInfo:
        Tgetter_table_column
        col = name_temp_flag
        tbl = name

    CommandLine:
        python -m wbia.control.manual_name_funcs --test-get_name_temp_flag

    RESTful:
        Method: GET
        URL:    /api/name/temp/

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.control.manual_name_funcs import *  # NOQA
        >>> ibs, config2_ = testdata_ibs()
        >>> name_rowid_list = ibs._get_all_name_rowids()
        >>> eager = True
        >>> name_temp_flag_list = ibs.get_name_temp_flag(name_rowid_list, eager=eager)
        >>> assert len(name_rowid_list) == len(name_temp_flag_list)
    """
    id_iter = name_rowid_list
    colnames = (NAME_TEMP_FLAG,)
    name_temp_flag_list = ibs.db.get(
        const.NAME_TABLE,
        colnames,
        id_iter,
        id_colname='rowid',
        eager=eager,
        nInput=nInput,
    )
    return name_temp_flag_list


@register_ibs_method
@register_api('/api/name/temp/', methods=['PUT'])
def set_name_temp_flag(
    ibs, name_rowid_list, name_temp_flag_list, duplicate_behavior='error'
):
    r"""
    name_temp_flag_list -> name.name_temp_flag[name_rowid_list]

    Args:
        name_rowid_list
        name_temp_flag_list

    TemplateInfo:
        Tsetter_native_column
        tbl = name
        col = name_temp_flag

    RESTful:
        Method: PUT
        URL:    /api/name/temp/
    """
    id_iter = name_rowid_list
    colnames = (NAME_TEMP_FLAG,)
    ibs.db.set(
        const.NAME_TABLE,
        colnames,
        name_temp_flag_list,
        id_iter,
        duplicate_behavior=duplicate_behavior,
    )


@register_ibs_method
@accessor_decors.getter_1to1
@register_api('/api/name/alias/text/', methods=['GET'], __api_plural_check__=False)
def get_name_alias_texts(ibs, name_rowid_list):
    r"""
    Returns:
        list_ (list): name_alias_text_list

    CommandLine:
        python -m wbia.control.manual_name_funcs --test-get_name_texts

    CommandLine:
        python -m wbia.control.manual_name_funcs --test-get_name_alias_texts

    RESTful:
        Method: GET
        URL:    /api/name/alias/text/

    Example:
        >>> # DISABLE_DOCTEST
        >>> from wbia.control.manual_name_funcs import *  # NOQA
        >>> import wbia
        >>> # build test data
        >>> ibs = wbia.opendb('testdb1')
        >>> name_rowid_list = ibs.get_valid_nids()
        >>> # execute function
        >>> name_alias_text_list = get_name_alias_texts(ibs, name_rowid_list)
        >>> # verify results
        >>> result = str(name_alias_text_list)
        >>> print(result)
        [None, None, None, None, None, None, None]
    """
    name_alias_text_list = ibs.db.get(
        const.NAME_TABLE, (NAME_ALIAS_TEXT,), name_rowid_list
    )
    return name_alias_text_list


@register_ibs_method
@accessor_decors.cache_invalidator(
    const.ANNOTATION_TABLE, [ANNOT_SEMANTIC_UUID], rowidx=None
)
@accessor_decors.setter
@register_api('/api/name/alias/text/', methods=['PUT'], __api_plural_check__=False)
def set_name_alias_texts(ibs, name_rowid_list, name_alias_text_list):
    r"""
    Returns:
        list_ (list): name_alias_text_list

    CommandLine:
        python -m wbia.control.manual_name_funcs --test-get_name_texts

    RESTful:
        Method: PUT
        URL:    /api/name/alias/text/
    """
    # ibsfuncs.assert_valid_names(name_alias_text_list)
    val_list = ((value,) for value in name_alias_text_list)
    ibs.db.set(const.NAME_TABLE, (NAME_ALIAS_TEXT,), val_list, name_rowid_list)
    # TODO: ibs.update_annot_semantic_uuids(aid_list)


@register_ibs_method
@accessor_decors.getter_1to1
@register_api('/api/name/text/', methods=['GET'])
def get_name_texts(ibs, name_rowid_list, apply_fix=True):
    r"""
    Returns:
        list_ (list): text names

    CommandLine:
        python -m wbia.control.manual_name_funcs --test-get_name_texts

    RESTful:
        Method: GET
        URL:    /api/name/text/

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.control.manual_name_funcs import *  # NOQA
        >>> import wbia
        >>> ibs = wbia.opendb('testdb1')
        >>> name_rowid_list = ibs._get_all_known_name_rowids()
        >>> name_text_list = get_name_texts(ibs, name_rowid_list)
        >>> result = ut.repr2(name_text_list)
        >>> print(result)
        ['easy', 'hard', 'jeff', 'lena', 'occl', 'polar', 'zebra']
    """
    # FIXME: Use standalone name table
    # TODO:
    # Change the temporary negative indexes back to the unknown NID for the
    # SQL query. Then augment the lblannot list to distinguish unknown lblannots
    # name_text_list = ibs.get_lblannot_values(nid_list, const.INDIVIDUAL_KEY)
    # name_text_list = ibs.get_lblannot_values(nid_list, const.INDIVIDUAL_KEY)
    name_text_list = ibs.db.get(const.NAME_TABLE, (NAME_TEXT,), name_rowid_list)
    if apply_fix:
        name_text_list = [
            const.UNKNOWN
            if rowid == const.UNKNOWN_NAME_ROWID or name_text is None
            else name_text
            for name_text, rowid in zip(name_text_list, name_rowid_list)
        ]
    return name_text_list


@register_ibs_method
def get_num_names(ibs, **kwargs):
    r"""
    Number of valid names

    CommandLine:
        python -m wbia.control.manual_name_funcs --test-get_num_names

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.control.manual_name_funcs import *  # NOQA
        >>> import wbia
        >>> ibs = wbia.opendb('testdb1')
        >>> result = get_num_names(ibs)
        >>> print(result)
        7
    """
    nid_list = ibs.get_valid_nids(**kwargs)
    return len(nid_list)


@register_ibs_method
@accessor_decors.getter_1to1
@register_api('/api/name/rowid/text/', methods=['GET'])
def get_name_rowids_from_text(ibs, name_text_list, ensure=True):
    r"""
    Args:
        ibs (IBEISController):  wbia controller object
        name_text_list (list):
        ensure (bool): adds as new names if non-existant (default = True)

    Returns:
        name_rowid_list (list): Creates one if it doesnt exist

    CommandLine:
        python -m wbia.control.manual_name_funcs --test-get_name_rowids_from_text:0
        python -m wbia.control.manual_name_funcs --test-get_name_rowids_from_text:1

    TODO:
        should ensure be defaulted to False?

    RESTful:
        Method: GET
        URL:    /api/name/rowid/text/

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.control.manual_name_funcs import *  # NOQA
        >>> import wbia
        >>> import utool as ut
        >>> ibs = wbia.opendb('testdb1')
        >>> name_text_list = [u'Fred', u'Sue', '____', u'zebra_grevys', 'TYPO', '____']
        >>> ensure = False
        >>> name_rowid_list = ibs.get_name_rowids_from_text(name_text_list, ensure)
        >>> print(ut.repr2(list(zip(name_text_list, name_rowid_list))))
        >>> ensure = True
        >>> name_rowid_list = ibs.get_name_rowids_from_text(name_text_list, ensure)
        >>> print(ut.repr2(list(zip(name_text_list, name_rowid_list))))
        >>> ibs.print_name_table()
        >>> result = str(name_rowid_list) + '\n'
        >>> typo_rowids = ibs.get_name_rowids_from_text(['TYPO', 'Fred', 'Sue', 'zebra_grevys'])
        >>> ibs.delete_names(typo_rowids)
        >>> result += str(ibs._get_all_known_name_rowids())
        >>> print('----')
        >>> ibs.print_name_table()
        >>> print(result)
        [8, 9, 0, 10, 11, 0]
        [1, 2, 3, 4, 5, 6, 7]
    """
    if ensure:
        name_rowid_list = ibs.add_names(name_text_list)
    else:
        name_rowid_list = ibs.get_name_rowids_from_text_(name_text_list)
    return name_rowid_list


@register_ibs_method
@accessor_decors.getter_1to1
def get_name_rowids_from_text_(ibs, name_text_list, ensure=True):
    r"""
    Args:
        ibs (IBEISController):  wbia controller object
        name_text_list (list):

    Returns:
        name_rowid_list (list):

    CommandLine:
        python -m wbia.control.manual_name_funcs --test-get_name_rowids_from_text_

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.control.manual_name_funcs import *  # NOQA
        >>> import wbia
        >>> import utool as ut  # NOQA
        >>> ibs = wbia.opendb('testdb1')
        >>> name_text_list = [u'Fred', 'easy', u'Sue', '____', u'zebra_grevys', 'TYPO', 'jeff']
        >>> name_rowid_list = ibs.get_name_rowids_from_text_(name_text_list)
        >>> ibs.print_name_table()
        >>> result = str(name_rowid_list)
        >>> print(result)
        [None, 1, None, 0, None, None, 3]
    """
    name_text_list_ = ibs.sanitize_name_texts(name_text_list)
    name_rowid_list = ibs.db.get(
        const.NAME_TABLE, (NAME_ROWID,), name_text_list_, id_colname=NAME_TEXT
    )
    name_rowid_list = [
        const.UNKNOWN_NAME_ROWID if text is None or text == const.UNKNOWN else rowid
        for rowid, text in zip(name_rowid_list, name_text_list_)
    ]
    return name_rowid_list


@register_ibs_method
@accessor_decors.getter_1to1
@register_api('/api/name/rowid/uuid/', methods=['GET'])
def get_name_rowids_from_uuid(ibs, uuid_list, nid_hack=False, ensure=True):
    r"""
    Args:
        ibs (IBEISController):  wbia controller object
        name_text_list (list):

    Returns:
        name_rowid_list (list):
    """
    name_rowid_list = ibs.db.get(
        const.NAME_TABLE, (NAME_ROWID,), uuid_list, id_colname=NAME_UUID
    )
    if nid_hack:
        name_rowid_list = [
            name_uuid if name_rowid is None else name_rowid
            for name_uuid, name_rowid in zip(uuid_list, name_rowid_list)
        ]
    return name_rowid_list


@register_ibs_method
@register_api('/api/name/dict/', methods=['GET'])
def get_name_nids_with_gids(ibs, nid_list=None):
    if nid_list is None:
        nid_list = sorted(ibs.get_valid_nids())
    name_list = ibs.get_name_texts(nid_list)
    gids_list = ibs.get_name_gids(nid_list)

    zipped = zip(nid_list, name_list, gids_list)
    combined_dict = {name: (nid, gid_list) for nid, name, gid_list in zipped}
    return combined_dict


@register_ibs_method
@accessor_decors.ider
@register_api('/api/name/', methods=['GET'])
def get_valid_nids(ibs, imgsetid=None, filter_empty=False, min_pername=None):
    r"""
    Returns:
        list_ (list): all valid names with at least one animal
        (does not include unknown names)

    RESTful:
        Method: GET
        URL:    /api/name/
    """
    if imgsetid is None:
        _nid_list = ibs._get_all_known_name_rowids()
    else:
        _nid_list = ibs.get_imageset_nids(imgsetid)
    # HACK FOR UNKNOWN. Makes things crash
    # _nid_list += [0]
    nid_list = _nid_list

    if filter_empty:
        min_pername = 1 if min_pername is None else max(min_pername, 1)

    if min_pername is not None:
        nAnnot_list = ibs.get_name_num_annotations(nid_list)
        flag_list = np.array(nAnnot_list) >= min_pername
        nid_list = ut.compress(nid_list, flag_list)
    return nid_list


@register_ibs_method
@accessor_decors.setter
@register_api('/api/name/note/', methods=['PUT'])
def set_name_notes(ibs, name_rowid_list, notes_list):
    r"""
    Sets a note for each name (multiple annotations)

    RESTful:
        Method: PUT
        URL:    /api/name/note/
    """
    # ibsfuncs.assert_lblannot_rowids_are_type(ibs, nid_list, ibs.lbltype_ids[const.INDIVIDUAL_KEY])
    # ibs.set_lblannot_notes(nid_list, notes_list)
    val_list = ((value,) for value in notes_list)
    ibs.db.set(const.NAME_TABLE, (NAME_NOTE,), val_list, name_rowid_list)


@register_ibs_method
@accessor_decors.setter
@register_api('/api/name/metadata/', methods=['PUT'])
def set_name_metadata(ibs, name_rowid_list, metadata_dict_list):
    r"""
    Sets the name's metadata using a metadata dictionary

    RESTful:
        Method: PUT
        URL:    /api/name/metadata/

    """
    id_iter = ((gid,) for gid in name_rowid_list)
    metadata_str_list = []
    for metadata_dict in metadata_dict_list:
        metadata_str = ut.to_json(metadata_dict)
        metadata_str_list.append(metadata_str)
    val_list = ((metadata_str,) for metadata_str in metadata_str_list)
    ibs.db.set(const.NAME_TABLE, ('name_metadata_json',), val_list, id_iter)


@register_ibs_method
@accessor_decors.setter
@register_api('/api/name/text/', methods=['PUT'])
def set_name_texts(
    ibs,
    name_rowid_list,
    name_text_list,
    verbose=False,
    notify_wildbook=False,
    assert_wildbook=False,
    update_json_log=True,
):
    r"""
    Changes the name text. Does not affect the animals of this name.
    Effectively just changes the TEXT UUID

    CommandLine:
        python -m wbia.control.manual_name_funcs --test-set_name_texts

    RESTful:
        Method: PUT
        URL:    /api/name/text/

    Example:
        >>> # DISABLE_DOCTEST
        >>> from wbia.control.manual_name_funcs import *  # NOQA
        >>> import wbia
        >>> ibs = wbia.opendb('testdb1')
        >>> nid_list = ibs.get_valid_nids()[0:2]
        >>> name_list = ibs.get_name_texts(nid_list)
        >>> result = set_name_texts(ibs, nid_list, name_list)
        >>> print(result)
    """
    import wbia

    if verbose:
        logger.info('[ibs] setting %d name texts' % (len(name_rowid_list),))
    if notify_wildbook and wbia.ENABLE_WILDBOOK_SIGNAL:
        logger.info('[ibs] notifying WildBook of name text changes')
        status_list = ibs.wildbook_signal_name_changes(name_rowid_list, name_text_list)

        wb_signaled = status_list is not None
        if assert_wildbook and wb_signaled:
            assert status_list, 'The request to WB failed'
            failed_nid_list = list(ut.ifilterfalse_items(name_rowid_list, status_list))
            args = (
                len(failed_nid_list),
                failed_nid_list,
            )
            msg = 'Failed to update %d WB names, nid_list = %r' % args
            assert len(failed_nid_list) == 0, msg
    ibsfuncs.assert_valid_names(name_text_list)
    old_name_text_list = ibs.get_name_texts(name_rowid_list)
    # sanitize_name_texts(ibs, name_text_list):
    # ibsfuncs.assert_lblannot_rowids_are_type(ibs, nid_list, ibs.lbltype_ids[const.INDIVIDUAL_KEY])
    # ibs.set_lblannot_values(nid_list, name_list)
    val_list = ((value,) for value in name_text_list)
    ibs.db.set(const.NAME_TABLE, (NAME_TEXT,), val_list, name_rowid_list)
    # Database updated, log name changes
    if update_json_log:
        import time

        json_log_path = ibs.get_logdir_local()
        json_log_filename = 'names.updates.json'
        json_log_filepath = os.path.join(json_log_path, json_log_filename)
        logger.info('Logging name changes to: %r' % (json_log_filepath,))
        # Log has never been made, create one
        if not os.path.exists(json_log_filepath):
            json_dict = {
                'updates': [],
            }
            json_str = ut.to_json(json_dict, pretty=True)
            with open(json_log_filepath, 'w') as json_log_file:
                json_log_file.write(json_str)
        # Get current log state
        with open(json_log_filepath, 'r') as json_log_file:
            json_str = json_log_file.read()
        json_dict = ut.from_json(json_str)
        db_name = ibs.get_db_name()
        db_init_uuid = ibs.get_db_init_uuid()
        # Zip all the updates together and write to updates list in dictionary
        zipped = zip(name_rowid_list, old_name_text_list, name_text_list)
        for name_rowid, old_name_text, new_name_text in zipped:
            json_dict['updates'].append(
                {
                    'time_unixtime': time.time(),
                    'db_name': db_name,
                    'db_init_uuid': db_init_uuid,
                    'name_rowid': name_rowid,
                    'name_old_text': old_name_text,
                    'name_new_text': new_name_text,
                }
            )
        # Write new log state
        json_str = ut.to_json(json_dict, pretty=True)
        with open(json_log_filepath, 'w') as json_log_file:
            json_log_file.write(json_str)


@register_ibs_method
@register_api('/api/name/sex/', methods=['GET'])
def get_name_sex(ibs, name_rowid_list, eager=True, nInput=None):
    r"""
    name_sex_list <- name.name_sex[name_rowid_list]

    gets data from the "native" column "name_sex" in the "name" table

    Args:
        name_rowid_list (list):

    Returns:
        list: name_sex_list

    TemplateInfo:
        Tgetter_table_column
        col = name_sex
        tbl = name

    RESTful:
        Method: GET
        URL:    /api/name/sex/

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.control.manual_name_funcs import *  # NOQA
        >>> ibs, config2_ = testdata_ibs()
        >>> name_rowid_list = ibs._get_all_name_rowids()
        >>> eager = True
        >>> name_sex_list = ibs.get_name_sex(name_rowid_list, eager=eager)
        >>> assert len(name_rowid_list) == len(name_sex_list)
    """
    id_iter = name_rowid_list
    colnames = (NAME_SEX,)
    name_sex_list = ibs.db.get(
        const.NAME_TABLE,
        colnames,
        id_iter,
        id_colname='rowid',
        eager=eager,
        nInput=nInput,
    )
    return name_sex_list


@register_ibs_method
@accessor_decors.setter
@register_api('/api/name/sex/', methods=['PUT'])
def set_name_sex(ibs, name_rowid_list, name_sex_list, duplicate_behavior='error'):
    r"""
    name_sex_list -> name.name_sex[name_rowid_list]

    Args:
        name_rowid_list
        name_sex_list

    TemplateInfo:
        Tsetter_native_column
        tbl = name
        col = name_sex

    RESTful:
        Method: PUT
        URL:    /api/name/sex/
    """
    id_iter = name_rowid_list
    colnames = (NAME_SEX,)
    ibs.db.set(
        const.NAME_TABLE,
        colnames,
        name_sex_list,
        id_iter,
        duplicate_behavior=duplicate_behavior,
    )


@register_ibs_method
@accessor_decors.getter_1to1
@register_api('/api/name/sex/text/', methods=['GET'])
def get_name_sex_text(ibs, name_rowid_list, eager=True, nInput=None):
    r"""

    RESTful:
        Method: GET
        URL:    /api/name/sex/text/
    """
    name_sex_list = ibs.get_name_sex(name_rowid_list, eager=eager, nInput=nInput)
    name_sex_text_list = ut.dict_take(const.SEX_INT_TO_TEXT, name_sex_list)
    return name_sex_text_list


@register_ibs_method
@accessor_decors.setter
@register_api('/api/name/sex/text/', methods=['PUT'])
def set_name_sex_text(ibs, name_rowid_list, name_sex_text_list):
    r"""

    RESTful:
        Method: PUT
        URL:    /api/name/sex/text/
    """
    name_sex_list = ut.dict_take(const.SEX_TEXT_TO_INT, name_sex_text_list)
    return ibs.set_name_sex(name_rowid_list, name_sex_list)


@register_ibs_method
@register_api('/api/name/age/months/min/', methods=['GET'], __api_plural_check__=False)
def get_name_age_months_est_min(ibs, name_rowid_list):
    r"""

    RESTful:
        Method: GET
        URL:    /api/name/age/months/min/
    """
    aids_list = ibs.get_name_aids(name_rowid_list)
    age_list = [ibs.get_annot_age_months_est_min(aid_list) for aid_list in aids_list]
    return age_list


@register_ibs_method
@register_api('/api/name/age/months/max/', methods=['GET'], __api_plural_check__=False)
def get_name_age_months_est_max(ibs, name_rowid_list):
    r"""

    RESTful:
        Method: GET
        URL:    /api/name/age/months/max/
    """
    aids_list = ibs.get_name_aids(name_rowid_list)
    age_list = [ibs.get_annot_age_months_est_max(aid_list) for aid_list in aids_list]
    return age_list


@register_ibs_method
@register_api('/api/name/imageset/rowid/', methods=['GET'])
def get_name_imgsetids(ibs, nid_list):
    r"""

    RESTful:
        Method: GET
        URL:    /api/name/imageset/rowid/
    """
    import utool as ut

    name_aids_list = ibs.get_name_aids(nid_list)
    name_aid_list = ut.flatten(name_aids_list)
    name_gid_list = ibs.get_annot_gids(name_aid_list)
    name_imgsetids_list = ibs.get_image_imgsetids(name_gid_list)
    name_imgsetid_list = ut.flatten(name_imgsetids_list)
    name_imgsetids = list(set(name_imgsetid_list))
    return name_imgsetids


@register_ibs_method
@register_api('/api/name/imageset/uuid/', methods=['GET'])
def get_name_imgset_uuids(ibs, nid_list):
    r"""
    RESTful:
        Method: GET
        URL:    /api/name/imageset/uuid/
    """
    name_imgsetids = ibs.get_name_imgsetids(nid_list)
    name_uuids_list = [
        ibs.get_imageset_uuids(name_imgsetid) for name_imgsetid in name_imgsetids
    ]
    return name_uuids_list


# def get_imageset_nids(ibs,


@register_ibs_method
@accessor_decors.getter
def get_name_has_split(ibs, nid_list):
    r"""
    CommandLine:
        python -m wbia.other.ibsfuncs --test-get_name_speeds

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.control.manual_name_funcs import *  # NOQA
        >>> import wbia
        >>> ibs = wbia.opendb('testdb1')
        >>> nid_list = ibs._get_all_known_nids()
        >>> splits_list = ibs.get_name_has_split(nid_list)
        >>> result = str(splits_list)
        >>> print(result)
    """
    aids_list_ = ibs.get_name_aids(nid_list)
    # ibs.check_name_mapping_consistency(aids_list_)

    def get_valid_aids_clique_annotmatch_rowids(aids):
        import itertools

        aid_pairs = list(itertools.combinations(aids, 2))
        aids1 = ut.take_column(aid_pairs, 0)
        aids2 = ut.take_column(aid_pairs, 1)
        am_ids = ibs.get_annotmatch_rowid_from_undirected_superkey(aids1, aids2)
        am_ids = ut.filter_Nones(am_ids)
        return am_ids

    amids_list = [get_valid_aids_clique_annotmatch_rowids(aids) for aids in aids_list_]
    flags_list = ibs.unflat_map(
        ut.partial(ibs.get_annotmatch_prop, 'SplitCase'), amids_list
    )
    has_splits = list(map(any, flags_list))
    return has_splits


@register_ibs_method
def get_name_speeds(ibs, nid_list):
    r"""
    CommandLine:
        python -m wbia.other.ibsfuncs --test-get_name_speeds

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.control.manual_name_funcs import *  # NOQA
        >>> import wbia
        >>> ibs = wbia.opendb('testdb1')
        >>> nid_list = ibs._get_all_known_nids()
        >>> speeds_list = get_name_speeds(ibs, nid_list)
        >>> result = str(speeds_list)
        >>> print(result)
    """
    aids_list = ibs.get_name_aids(nid_list)
    # ibs.check_name_mapping_consistency(aids_list_)
    # aids_list = [(aids) for aids in aids_list_]
    # speeds_list = ibs.get_unflat_annots_speeds_list(aids_list)
    speeds_list = ibs.get_unflat_annots_speeds_list2(aids_list)
    return speeds_list


@register_ibs_method
@accessor_decors.getter
def get_name_hourdiffs(ibs, nid_list):
    """
    CommandLine:
        python -m wbia.other.ibsfuncs --test-get_name_hourdiffs

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.control.manual_name_funcs import *  # NOQA
        >>> import wbia
        >>> ibs = wbia.opendb('testdb1')
        >>> nid_list = ibs._get_all_known_nids()
        >>> hourdiffs_list = ibs.get_name_hourdiffs(nid_list)
        >>> result = hourdiffs_list
        >>> print(hourdiffs_list)
    """
    aids_list_ = ibs.get_name_aids(nid_list)
    # ibs.check_name_mapping_consistency(aids_list_)
    # HACK FILTERING SHOULD NOT OCCUR HERE
    aids_list = [(aids) for aids in aids_list_]
    hourdiffs_list = ibs.get_unflat_annots_hourdists_list(aids_list)
    return hourdiffs_list


@register_ibs_method
@accessor_decors.getter
def get_name_max_hourdiff(ibs, nid_list):
    hourdiffs_list = ibs.get_name_hourdiffs(nid_list)
    maxhourdiff_list_ = np.array(
        [vt.safe_max(hourdiff, nans=False) for hourdiff in hourdiffs_list]
    )
    maxhourdiff_list = np.array(maxhourdiff_list_)
    return maxhourdiff_list


@register_ibs_method
@accessor_decors.getter
def get_name_max_speed(ibs, nid_list):
    """
    CommandLine:
        python -m wbia.other.ibsfuncs --test-get_name_max_speed

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.control.manual_name_funcs import *  # NOQA
        >>> import wbia
        >>> ibs = wbia.opendb('testdb1')
        >>> nid_list = ibs._get_all_known_nids()
        >>> maxspeed_list = ibs.get_name_max_speed(nid_list)
        >>> result = maxspeed_list
        >>> print(maxspeed_list)
    """
    speeds_list = ibs.get_name_speeds(nid_list)
    maxspeed_list = np.array([vt.safe_max(speeds, nans=False) for speeds in speeds_list])
    return maxspeed_list


@register_ibs_method
def get_name_gps_tracks(ibs, nid_list=None, aid_list=None):
    """
    CommandLine:
        python -m wbia.other.ibsfuncs --test-get_name_gps_tracks

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.control.manual_name_funcs import *  # NOQA
        >>> import wbia
        >>> # build test data
        >>> #ibs = wbia.opendb('PZ_Master0')
        >>> ibs = wbia.opendb('testdb1')
        >>> #nid_list = ibs.get_valid_nids()
        >>> aid_list = ibs.get_valid_aids()
        >>> nid_list, gps_track_list, aid_track_list = ibs.get_name_gps_tracks(aid_list=aid_list)
        >>> nonempty_list = list(map(lambda x: len(x) > 0, gps_track_list))
        >>> ut.compress(nid_list, nonempty_list)
        >>> ut.compress(gps_track_list, nonempty_list)
        >>> ut.compress(aid_track_list, nonempty_list)
        >>> aid_track_list = list(map(sorted, aid_track_list))
        >>> result = str(aid_track_list)
        >>> print(result)
        [[11], [], [4], [1], [2, 3], [5, 6], [7], [8], [10], [12], [13]]
    """
    assert aid_list is None or nid_list is None, 'only specify one please'
    if aid_list is None:
        aids_list_ = ibs.get_name_aids(nid_list)
    else:
        aids_list_, nid_list = ibs.group_annots_by_name(aid_list)
    aids_list = [
        ut.sortedby(aids, ibs.get_annot_image_unixtimes(aids)) for aids in aids_list_
    ]
    gids_list = ibs.unflat_map(ibs.get_annot_gids, aids_list)
    gpss_list = ibs.unflat_map(ibs.get_image_gps, gids_list)

    isvalids_list = [
        [gps[0] != -1.0 or gps[1] != -1.0 for gps in gpss] for gpss in gpss_list
    ]
    gps_track_list = [
        ut.compress(gpss, isvalids) for gpss, isvalids in zip(gpss_list, isvalids_list)
    ]
    aid_track_list = [
        ut.compress(aids, isvalids) for aids, isvalids in zip(aids_list, isvalids_list)
    ]
    return nid_list, gps_track_list, aid_track_list



# -*- coding: utf-8 -*-
# class FitHarnessOld(object):
#     def __init__(harn, train_loader, vali_loader, test_loader):
#         harn.train_loader = train_loader
#         harn.vali_loader = vali_loader
#         harn.test_loader = test_loader
#         harn.criterion = ContrastiveLoss(margin=1.0)
#         harn.model = Siamese()
#         harn.lr_scheduler = LRSchedule.exp
#         harn.use_cuda = False
#         # harn.model = torch.nn.DataParallel(model, device_ids=[0, 1]).cuda()
#         harn.config = {
#             'maxIterations': 10000,
#             'displayInterval': 1,
#             'vail_displayInterval': 1,
#             'model_dir': '.',
#             'margin': 1.0,
#         }
#         harn.lr = 0.001
#         harn.epoch = 0

#     def log(harn, msg):
#         print(msg)

#     def log_value(harn, key, value, n_iter):
#         print('{}={} @ {}'.format(key, value, n_iter))

#     def load_snapshot(harn, load_path):
#         snapshot = torch.load(load_path)
#         # loadModelState(model, snapshot)
#         harn.model.load_state_dict(snapshot['state_dict'])
#         harn.epoch = snapshot['epoch'] + 1
#         harn.log('Model loaded from {}'.format(load_path))

#     def run(harn):
#         # optimizer = harn.config.optimizer(model.parameters(), lr=lr)
#         # harn.optimizer = torch.optim.SGD(harn.model.parameters(), lr=harn.lr)
#         harn.optimizer = torch.optim.Adam(harn.model.parameters(), lr=harn.lr)

#         # train loop
#         # configure("runs/afrl", flush_secs=2)

#         while True:
#             harn.train_epoch()
#             harn.validation_epoch()
#             harn.save_snapshot()

#             # check for termination
#             if harn.epoch > harn.config['maxIterations']:
#                 harn.log('Maximum harn.epoch reached, terminating ...')
#                 break
#             harn.epoch += 1

#     def train_epoch(harn):
#         ave_metrics = {
#             'loss': 0,
#             'accuracy': 0,
#             'pos_dist': 0,
#             'neg_dist': 0,
#         }

#         # change learning rate
#         harn.optimizer, harn.lr = harn.lr_scheduler(harn.optimizer, harn.epoch, harn.lr, 2)

#         # train batch
#         for batch_idx, (data0, data1, target) in enumerate(harn.train_loader):
#             target = target.type(torch.FloatTensor)
#             if harn.use_cuda:
#                 data0, data1, target = data0.cuda(), data1.cuda(), target.cuda()
#             data0, data1, target = Variable(data0), Variable(data1), Variable(target)
#             input_batch = (data0, data1, target)
#             print('Begin batch {}'.format(batch_idx))
#             t_cur_metrics = harn.train_batch(input_batch)

#             for k, v in t_cur_metrics.items():
#                 ave_metrics[k] += v

#             # display training info
#             if (batch_idx + 1) % harn.config['displayInterval'] == 0:
#                 for k in ave_metrics.keys():
#                     ave_metrics[k] /= harn.config['displayInterval']

#                 n_train = len(harn.train_loader)
#                 harn.log('Epoch {0}: {1} / {2} | lr:{3} - tloss:{4:.5f} acc:{5:.2f} | sdis:{6:.3f} ddis:{7:.3f}'.format(
#                     harn.epoch, batch_idx, n_train, harn.lr,
#                     ave_metrics['loss'], ave_metrics['accuracy'],
#                     ave_metrics['pos_dist'], ave_metrics['neg_dist']))

#                 iter_idx = harn.epoch * n_train + batch_idx
#                 for key, value in ave_metrics.items():
#                     harn.log_value('train ' + key, value, iter_idx)

#                 # diagnoseGradients(model.parameters())
#                 for k in ave_metrics.keys():
#                     ave_metrics[k] = 0

#     def validation_epoch(harn):
#         ave_metrics = {
#             'loss': 0,
#             'accuracy': 0,
#             'pos_dist': 0,
#             'neg_dist': 0,
#         }

#         final_metrics = ave_metrics.copy()

#         for vali_idx, (t_data0, t_data1, t_target) in enumerate(harn.vali_loader):
#             t_target = t_target.type(torch.FloatTensor)
#             if harn.use_cuda:
#                 t_data0, t_data1, t_target = t_data0.cuda(), t_data1.cuda(), t_target.cuda()
#             t_data0, t_data1, t_target = Variable(t_data0), Variable(t_data1), Variable(t_target)

#             input_batch = (t_data0, t_data1, t_target)
#             print('Begin batch {}'.format(vali_idx))
#             v_cur_metrics = harn.validation_batch(input_batch)

#             for k, v in v_cur_metrics.items():
#                 ave_metrics[k] += v
#                 final_metrics[k] += v

#             if (vali_idx + 1) % harn.config['vail_displayInterval'] == 0:
#                 for k in ave_metrics.keys():
#                     ave_metrics[k] /= harn.config['displayInterval']

#                 harn.log('Epoch {0}: {1} / {2} | vloss:{3:.5f} acc:{4:.2f} | sdis:{5:.3f} ddis:{6:.3f}'.format(
#                     harn.epoch, vali_idx, len(harn.vali_loader),
#                     ave_metrics['loss'], ave_metrics['accuracy'],
#                     ave_metrics['pos_dist'], ave_metrics['neg_dist']))

#                 for k in ave_metrics.keys():
#                     ave_metrics[k] = 0

#         for k in final_metrics.keys():
#             final_metrics[k] /= len(harn.vali_loader)
#         harn.log('Epoch {0}: final vloss:{1:.5f} acc:{2:.2f} | sdis:{3:.3f} ddis:{4:.3f}'.format(
#             harn.epoch, final_metrics['loss'], final_metrics['accuracy'],
#             final_metrics['pos_dist'], final_metrics['neg_dist']))

#         iter_idx = harn.epoch * len(harn.vali_loader) + vali_idx
#         for key, value in final_metrics.items():
#             harn.log_value('validation ' + key, value, iter_idx)

#     def save_snapshot(harn):
#         # save snapshot
#         save_path = join(harn.config['model_dir'], 'snapshot_epoch_{}.pt'.format(harn.epoch))
#         # torch.save(checkpoint(model, harn.epoch), save_path)
#         harn.log('Snapshot saved to {}'.format(save_path))

#     def train_batch(harn, input_batch):
#         harn.model.train(True)
#         input1, input2, label = input_batch
#         output0, output1, output = harn.model(input1, input2)
#         t_metrics = harn._measure_metrics(output0, output1, label, harn.config['margin'])

#         # loss = harn.criterion(output, label)
#         loss = harn.criterion(output0, output1, label)

#         harn.optimizer.zero_grad()
#         loss.backward()
#         harn.optimizer.step()

#         # loss = loss / input1.size()[0]

#         loss_sum = loss.data.sum()

#         inf = float("inf")
#         if loss_sum == inf or loss_sum == -inf:
#             harn.log("WARNING: received an inf loss, setting loss value to 0")
#             loss_value = 0
#         else:
#             loss_value = loss.data[0]

#         t_metrics['loss'] = loss_value
#         return t_metrics

#     def validation_batch(harn, input_batch):
#         harn.model.train(False)
#         input1, input2, label = input_batch

#         output0, output1, output = harn.model(input1, input2)
#         v_metrics = harn._measure_metrics(output0, output1, label, harn.config['margin'])

#         # loss = harn.criterion(output, label)
#         loss = harn.criterion(output0, output1, label)
#         # loss = loss / input1.size()[0]
#         loss_sum = loss.data.sum()

#         inf = float("inf")
#         if loss_sum == inf or loss_sum == -inf:
#             harn.log("WARNING: received an inf loss, setting loss value to 0")
#             loss_value = 0
#         else:
#             loss_value = loss.data[0]

#         v_metrics['loss'] = loss_value
#         return v_metrics

#     def _measure_metrics(harn, output0, output1, label, margin):
#         diff = torch.abs(output0 - output1)
#         l21 = torch.sqrt(torch.pow(diff, 2).sum(dim=1))

#         label_tensor = torch.from_numpy(label.data.cpu().numpy())
#         l21_tensor = torch.from_numpy(l21.data.cpu().numpy())

#         # Distance
#         is_pos = torch.ByteTensor()
#         POS_LABEL = 1
#         NEG_LABEL = 0
#         torch.eq(label_tensor, POS_LABEL, out=is_pos)  # y==1
#         pos_dist = 0 if len(l21_tensor[is_pos]) == 0 else l21_tensor[is_pos].mean()
#         neg_dist = 0 if len(l21_tensor[~is_pos]) == 0 else l21_tensor[~is_pos].mean()
#         # print('same dis : diff dis  {} : {}'.format(l21_tensor[is_pos == 0].mean(), l21_tensor[is_pos].mean()))

#         # accuracy
#         pred_pos_flags = torch.ByteTensor()
#         torch.le(l21_tensor, margin, out=pred_pos_flags)  # y==1's idx

#         cur_score = torch.FloatTensor(label.size(0))
#         cur_score.fill_(NEG_LABEL)
#         cur_score[pred_pos_flags] = POS_LABEL

#         accuracy = torch.eq(cur_score, label_tensor).sum() / label_tensor.size(0)

#         metrics = {
#             'accuracy': accuracy,
#             'pos_dist': pos_dist,
#             'neg_dist': neg_dist,
#         }

#         return metrics
